﻿using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Services.Authentication.DataAccess;
using DA = HeuristicLab.Services.Authentication.DataAccess;
using DT = HeuristicLab.Services.Authentication.DataTransfer;


namespace HeuristicLab.Services.Authentication {

  public class AuthenticationService : IAuthenticationService {

    public DataTransfer.Client GetClient(Guid id) {
      using (ClientManagmentDataContext cmdc = new ClientManagmentDataContext()) {
        DT.Client client = Convert.ToDto(cmdc.Resources.OfType<DA.Client>().FirstOrDefault(x => x.Id == id));
        if (client != null) {
          var groups = cmdc.ResourceResourceGroups.Where(x => x.ResourceId == id).Select(x => x.ResourceGroupId).ToArray();
          client.ResourceGroup = groups;
        }
        return client;
      }
    }

    public IEnumerable<DataTransfer.Client> GetClients() {
      using (ClientManagmentDataContext cmdc = new ClientManagmentDataContext()) {
        List<DT.Client> allClients = new List<DT.Client>();
        foreach (DA.Client dbClient in cmdc.Resources.OfType<DA.Client>()) {
          DT.Client client = Convert.ToDto(dbClient);
          ////////if (client != null) {
          ////////  DA.ResourceResourceGroup rrg = cmdc.ResourceResourceGroups.FirstOrDefault(x => x.ResourceId == client.Id);
          ////////  if (rrg != null)
          ////////    //client.ResourceGroup. = rrg.ResourceGroupId;
          ////////  allClients.Add(client);
          ////////}
          if (client != null) {
            var groups = cmdc.ResourceResourceGroups.Where(x => x.ResourceId == client.Id).Select(x => x.ResourceGroupId).ToArray();
            client.ResourceGroup = groups;
          }
          allClients.Add(client);
        }

        if (allClients.Count > 0)
          return allClients;
        return null;
      }
    }

    public Guid AddClient(DataTransfer.Client dto) {
      using (ClientManagmentDataContext cmdc = new ClientManagmentDataContext()) {
        return InsertResource(cmdc, dto);
      }
    }


    public void UpdateClient(DataTransfer.Client dto) {
      using (ClientManagmentDataContext cmdc = new ClientManagmentDataContext()) {
        DA.Client client = cmdc.Resources.OfType<DA.Client>().FirstOrDefault(x => x.Id == dto.Id);

        // check if resource exists
        if (client == null)
          return;

        UpdateResourceResourceGroup(cmdc, dto);

        DA.Client entity = Convert.ToEntity(dto);
        //entity.Id = dto.Id;
        //entity.ResourceType = dto.ResourceType;
        client.Name = entity.Name;
        client.ResourceGroups = entity.ResourceGroups;
        client.Description = entity.Description;
        client.IPAdress = entity.IPAdress;
        client.MemorySize = entity.MemorySize;
        client.NumberOfProcessors = entity.NumberOfProcessors;
        client.NumberOfThreads = entity.NumberOfThreads;
        client.OperatingSystem = entity.OperatingSystem;
        client.ProcessorType = entity.ProcessorType;

        cmdc.SubmitChanges();
      }

    }

    public void DeleteClient(DataTransfer.Client dto) {
      using (ClientManagmentDataContext cmdc = new ClientManagmentDataContext()) {
        RemoveResource(cmdc, dto);
      }
    }

    public DataTransfer.ResourceGroup GetResourceGroup(Guid id) {
      using (ClientManagmentDataContext cmdc = new ClientManagmentDataContext()) {
        DA.ResourceGroup resourceGroup = cmdc.Resources.OfType<DA.ResourceGroup>().FirstOrDefault(x => x.Id == id);
        return Convert.ToDto(resourceGroup);
      }
    }

    public IEnumerable<DataTransfer.ResourceGroup> GetResourceGroupsOf(Guid id) {
      List<DT.ResourceGroup> allResourceGroups = new List<DT.ResourceGroup>();
      using (ClientManagmentDataContext cmdc = new ClientManagmentDataContext()) {
        var groups = cmdc.ResourceResourceGroups.Where(x => x.ResourceId == id).Select(x => x.ResourceGroupId).ToArray();
        foreach (Guid rId in groups)
          allResourceGroups.Add(GetResourceGroup(rId));
        return allResourceGroups;
      }
    }

    public IEnumerable<DataTransfer.ResourceGroup> GetResourceGroups() {
      using (ClientManagmentDataContext cmdc = new ClientManagmentDataContext()) {
        List<DT.ResourceGroup> allClients = new List<DT.ResourceGroup>();
        foreach (DA.ResourceGroup dbResourceGroup in cmdc.Resources.OfType<DA.ResourceGroup>()) {
          DT.ResourceGroup resourceGroup = Convert.ToDto(dbResourceGroup);
          if (resourceGroup != null)
            allClients.Add(resourceGroup);
        }
        if (allClients.Count > 0)
          return allClients;
        return null;
      }
    }

    public Guid AddResourceGroup(DataTransfer.ResourceGroup dto) {
      using (ClientManagmentDataContext cmdc = new ClientManagmentDataContext()) {
        return InsertResource(cmdc, dto);
      }
    }

    public void UpdateResourceGroup(DataTransfer.ResourceGroup dto) {
      using (ClientManagmentDataContext cmdc = new ClientManagmentDataContext()) {
        DA.ResourceGroup group = cmdc.Resources.OfType<DA.ResourceGroup>().FirstOrDefault(x => x.Id == dto.Id);

        // check if resource exists
        if (group == null)
          return;
        
        UpdateResourceResourceGroup(cmdc, dto);

        DA.ResourceGroup entity = Convert.ToEntity(dto);
        ////group.Id;
        ////group.ResourceType;
        group.Name = entity.Name;
        //group.ResourceGroups = entity.ResourceGroups;
        group.Description = entity.Description;

        cmdc.SubmitChanges();
      }
    }

    public void DeleteResourceGroup(DataTransfer.ResourceGroup dto) {
      using (ClientManagmentDataContext cmdc = new ClientManagmentDataContext()) {
        RemoveResource(cmdc, dto);
      }
    }

    // neu
    public IEnumerable<DataTransfer.Resource> GetMembers(Guid resourceGroupid) {
      List<DT.Resource> allMembers = new List<DT.Resource>();
      using (ClientManagmentDataContext cmdc = new ClientManagmentDataContext()) {
        var members = cmdc.ResourceResourceGroups.Where(x => x.ResourceGroupId == resourceGroupid).Select(x => x.ResourceId).ToArray();
        foreach (Guid rId in members)
          allMembers.Add(GetClient(resourceGroupid));
        return allMembers;
      }
    }

    public IEnumerable<DataTransfer.Resource> GetRootResources() {
      List<DT.ResourceGroup> rootResources = new List<DT.ResourceGroup>();
      using (ClientManagmentDataContext cmdc = new ClientManagmentDataContext()) {
        foreach (DA.ResourceGroup dbResourceGroup in cmdc.Resources.OfType<DA.ResourceGroup>()) {
          var rrg = cmdc.ResourceResourceGroups.Where(x => x.ResourceId == dbResourceGroup.Id);
          if ((rrg.Count() == 0) && 
            (dbResourceGroup.ResourceType == DT.ResourceGroup.ResourceTypes.ResourceGroup.ToString())) {
            DT.ResourceGroup group = Convert.ToDto(dbResourceGroup);
            rootResources.Add(group);
          }
        }
        return rootResources;
      }
    }

    private Guid InsertResource(ClientManagmentDataContext cmdc, DT.Resource dto) {
      List<ResourceResourceGroup> newRrgList = new List<ResourceResourceGroup>();
      if ((dto.ResourceGroup != null) && (dto.ResourceGroup.Count() > 0)) {
        foreach (Guid rgId in dto.ResourceGroup) {
          ResourceGroup grp = cmdc.Resources.OfType<DA.ResourceGroup>().FirstOrDefault(x => x.Id == rgId);
          if (grp == null)
            return Guid.Empty;
          ResourceResourceGroup rrg = new ResourceResourceGroup();
          rrg.ResourceGroup = grp;
          newRrgList.Add(rrg);
        }
      }
      DA.Resource entity = Convert.ToEntity(dto);
      if (entity != null) {
        entity.Id = Guid.NewGuid();
        cmdc.Resources.InsertOnSubmit(entity);
        if (newRrgList.Count() > 0) {
          foreach (ResourceResourceGroup rrg in newRrgList) {
            rrg.Client = entity;
            cmdc.ResourceResourceGroups.InsertOnSubmit(rrg);
          }
        }
        cmdc.SubmitChanges();
        return entity.Id;
      }
      return Guid.Empty;
    }

    private static bool RemoveResource(ClientManagmentDataContext cmdc, DT.Resource dto) {
      RemoveRRGEntry(cmdc, dto.Id);
      DA.Resource entity = cmdc.Resources.FirstOrDefault(x => x.Id == dto.Id);
      if (entity != null) {
        cmdc.Resources.DeleteOnSubmit(entity);
        cmdc.SubmitChanges();
        return true;
      }
      return false;
    }

    private static void RemoveRRGEntry(ClientManagmentDataContext cmdc, Guid resID) {
      foreach (var rrg in cmdc.ResourceResourceGroups.Where(x => x.ResourceId == resID))
        cmdc.ResourceResourceGroups.DeleteOnSubmit(rrg);
      foreach (var rrg in cmdc.ResourceResourceGroups.Where(x => x.ResourceGroupId == resID)) {
        // Begin Cascade Delete
        RemoveRRGEntry(cmdc, rrg.Client.Id);
        cmdc.Resources.DeleteOnSubmit(rrg.Client);
        // End Cascade Delete
        cmdc.ResourceResourceGroups.DeleteOnSubmit(rrg);
      }
    }

    private void UpdateResourceResourceGroup(ClientManagmentDataContext cmdc, DataTransfer.Resource dto) {
      //delete if no more memeber

      var rrgList = cmdc.ResourceResourceGroups.Where(x => x.ResourceId == dto.Id);
      foreach (ResourceResourceGroup rrg in rrgList) {
        if (!dto.ResourceGroup.Contains(rrg.ResourceGroupId))
          cmdc.ResourceResourceGroups.DeleteOnSubmit(rrg);
      }

      // insert new rrg if rrg-entry not exists
      var rrgResourceGroupIdList = rrgList.Select(x => x.ResourceGroupId).ToArray();
      foreach (Guid rrgResGroupId in dto.ResourceGroup) {
        if (!rrgResourceGroupIdList.Contains(rrgResGroupId)) {
          ResourceGroup grp = cmdc.Resources.OfType<DA.ResourceGroup>().FirstOrDefault(x => x.Id == rrgResGroupId);
          if (grp == null)
            return;
          ResourceResourceGroup newRrg = new ResourceResourceGroup();
          newRrg.ResourceId = dto.Id;
          newRrg.ResourceGroupId = grp.Id; //=rrgResGroupId
          cmdc.ResourceResourceGroups.InsertOnSubmit(newRrg);
        }
      }
      cmdc.SubmitChanges();
    }

  }
}